/*
 * This program is intended to "unload" a Citadel 5 system into the
 * intermediate format used by Citadel 6.
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "citadel.h"

#define V_FORGET        1               /* User has zapped this room        */
#define V_LOCKOUT       2               /* User is locked out of this room  */
#define V_ACCESS        4               /* Access is granted to this room   */


struct config config;
struct msgmain mm;
char roomnames[MAXROOMS][20];
long roomnums[MAXROOMS];
char roomgens[MAXROOMS];
long highest_roomnum = 0L;

void pwcrypt(char *text, int code)
{
        int a;
        for (a=0; a<strlen(text); ++a) text[a]=(text[a]^(((code|128)^a)&0xFF));
        }


/*
 * Read into a temporary file...
 */
int readmessage(char *tempname, FILE *sourcefp, long pos) {
	FILE *destfp;
	int ch, field;

	destfp = fopen(tempname, "wb");
	if (destfp == NULL) {
		perror("error opening temp file");
		return(errno);
		}

	fseek(sourcefp, pos, 0);
	ch = getc(sourcefp);
	if (ch != 255) {
		printf("*** can't locate message!  pos=%ld\n", pos);
		fclose(destfp);
		return(1);
		}
	putc(ch, destfp);

	ch = getc(sourcefp);	putc(ch, destfp);	/* type */
	ch = getc(sourcefp);	putc(ch, destfp);	/* format */
	
	do {
		field = getc(sourcefp);
		putc(field, destfp);
		if ( (field != 'M') && (field > 0) ) {
			do {
				ch = getc(sourcefp);
				putc(ch, destfp);
				} while (ch > 0);
			}
		} while ( (field != 'M') && (field > 0) );

	if (field != 'M') {
		printf("*** truncated message\n");
		putc(0,destfp);
		fclose(destfp);
		return(2);
		}

	while (ch = getc(sourcefp), (ch > 0)) {
		putc(ch, destfp);
		}
	putc(0, destfp);
	fclose(destfp);
	return(0);
	}






void dump_message(FILE *exfp, long msgnum, long msgpos) {
	static FILE *msg = NULL;
	FILE *tmp;
	char tempfilename[64];
	long count, bytes;
	int ch;

	if (msg == NULL) msg = fopen("msgmain", "rb");
	sprintf(tempfilename, "/tmp/export5.%d", getpid());
	readmessage(tempfilename, msg, msgpos);
	tmp = fopen(tempfilename, "rb");
	fseek(tmp, 0L, 2);
	bytes = ftell(tmp);
	fprintf(exfp, "%ld%c", bytes, 0);
	fseek(tmp, 0L, 0);
	for (count=0L; count<bytes; ++count) {
		ch = getc(tmp);
		putc(ch, exfp);
		}
	fclose(tmp);
	unlink(tempfilename);
	}


void export_usersupp(FILE *exfp, FILE *vifp) {
	FILE *fp;
	struct usersupp us;
	int a,f;

	fp = fopen("usersupp", "rb+");
	if (fp == NULL) {
		perror("Cannot open usersupp");
		fclose(exfp);
		exit(1);
		}

	printf("\n\nUsers\n\n");
	fprintf(exfp, "usersupp%c", 0);

	while (fread(&us, sizeof(struct usersupp), 1, fp) > 0)
	    if (us.axlevel > 0) {
		printf("User <%s>", us.fullname);
		fflush(stdout);
		fprintf(exfp, "user%c", 0);

		fprintf(exfp, "usuid%c%d%c", 0, us.USuid, 0);
		pwcrypt(us.password, PWCRYPT);
		fprintf(exfp, "password%c%s%c", 0, us.password, 0);

		fprintf(exfp, "flags%c%d%c", 0, us.flags, 0);
		fprintf(exfp, "timescalled%c%d%c", 0, us.timescalled, 0);
		fprintf(exfp, "posted%c%d%c", 0, us.posted, 0);
		fprintf(exfp, "fullname%c%s%c", 0, us.fullname, 0);
		fprintf(exfp, "axlevel%c%d%c", 0, us.axlevel, 0);
		fprintf(exfp, "usscreenwidth%c%d%c", 0, us.USscreenwidth, 0);
		fprintf(exfp, "usscreenheight%c%d%c", 0, us.USscreenheight, 0);
		fprintf(exfp, "usernum%c%ld%c", 0, us.eternal, 0);
		fprintf(exfp, "lastcall%c%ld%c", 0, us.lastcall, 0);
		fprintf(exfp, "usname%c%s%c", 0, us.USname, 0);
		fprintf(exfp, "usaddr%c%s%c", 0, us.USaddr, 0);
		fprintf(exfp, "uscity%c%s%c", 0, us.UScity, 0);
		fprintf(exfp, "usstate%c%s%c", 0, us.USstate, 0);
		fprintf(exfp, "uszip%c%s%c", 0, us.USzip, 0);
		fprintf(exfp, "usphone%c%s%c", 0, us.USphone, 0);
		fprintf(exfp, "usemail%c%s%c", 0, us.USemail, 0);

		/* Translate Cit5 generation stuff into Cit6 'visit' struct */
		for (a=0; a<MAXROOMS; ++a)
		    if ( (roomgens[a] >= 0) 
		      && (a != 1)
		      && ( (us.generation[a]>=0) || (us.forget[a]>=0) )
		      ) {
			fprintf(vifp, "visit%c", 0);
			fprintf(vifp, "vrnum%c%ld%c", 0, roomnums[a], 0);
			fprintf(vifp, "vgen%c%d%c", 0, roomgens[a], 0);
			fprintf(vifp, "vunum%c%ld%c", 0, us.eternal, 0);
			fprintf(vifp, "lastseen%c%d%c", 0, us.lastseen[a], 0);
			f = 0;
			if (us.generation[a]==roomgens[a]) f = f | V_ACCESS;
			if (us.forget[a]==roomgens[a]) f = f | V_FORGET;
			fprintf(vifp, "flags%c%d%c", 0, f, 0);
			fprintf(vifp, "endvisit%c", 0);
			}

		/********** we don't export mail room visits; it's too hard
		fprintf(exfp, "visit%c", 0);
		fprintf(exfp, "vname%c%010ld.Mail%c", 0, us.eternal, 0);
		fprintf(exfp, "vgen%c1%c", 0, 0);	* always 1 *
		fprintf(exfp, "lastseen%c%d%c", 0, us.lastseen[1], 0);
		f = V_ACCESS;
		fprintf(exfp, "flags%c%d%c", 0, f, 0);
		fprintf(exfp, "endvisit%c", 0);
		*********************************************************/

		printf("\n");

		fprintf(exfp, "enduser%c", 0);
		}

	fclose(fp);
	fprintf(exfp, "endsection%c", 0);
	}
	

void export_floors(FILE *exfp) {
	FILE *flfp;
	int floornum;
	struct floor fl;

	flfp = fopen("floortab", "rb");
	fprintf(exfp, "floors%c", 0);
	if (flfp == NULL) {
		perror("cannot open floortab");
		exit(1);
		}
	for (floornum=0; floornum<MAXFLOORS; ++floornum) {
		fread(&fl, sizeof(struct floor), 1, flfp);
		fprintf(exfp, "floor%c", 0);
		fprintf(exfp, "f_flags%c%d%c", 0, fl.f_flags, 0);
		fprintf(exfp, "f_name%c%s%c", 0, fl.f_name, 0);
		fprintf(exfp, "f_ref_count%c%d%c", 0, fl.f_ref_count, 0);
		fprintf(exfp, "endfloor%c", 0);
		}

	fclose(flfp);
	fprintf(exfp, "endsection%c", 0);

	}


void export_rooms(FILE *exfp) {
	FILE *qrfp, *frfp, *usfp;
	char frname[100];
	struct quickroom qr;
	struct fullroom fr;
	struct usersupp us;
	int a, b;
	int msgcount = 0;
	long mailroomnum;

	printf("\n\nRooms\n\n");
	qrfp = fopen("quickroom", "rb");
	if (qrfp == NULL) {
		perror("cannot open quickroom");
		exit(1);
		}

	fprintf(exfp, "rooms%c", 0);
	for (a=0; a<MAXROOMS; ++a) {
	    roomnums[a] = highest_roomnum++;
	    msgcount = 0;
	    printf("Room #%3d ", a);
	    fflush(stdout);
	    fread(&qr, sizeof(struct quickroom), 1, qrfp);
	    strcpy(roomnames[a], qr.QRname);
	    roomgens[a] = (-1);
	    if (qr.QRflags & QR_INUSE) {
		printf("<%s><%ld> ", qr.QRname, roomnums[a]);
		fflush(stdout);
	        roomgens[a] = qr.QRgen;
		fprintf(exfp, "room%c", 0);
		fprintf(exfp, "qrname%c%s%c", 0, qr.QRname, 0);
		fprintf(exfp, "qrpasswd%c%s%c", 0, qr.QRpasswd, 0);
		fprintf(exfp, "qrroomaide%c%ld%c", 0, qr.QRroomaide, 0);
		fprintf(exfp, "qrhighest%c%ld%c", 0, qr.QRhighest, 0);
		fprintf(exfp, "qrgen%c%d%c", 0, qr.QRgen, 0);
		fprintf(exfp, "qrflags%c%d%c", 0, qr.QRflags, 0);
		fprintf(exfp, "qrdirname%c%s%c", 0, qr.QRdirname, 0);
		fprintf(exfp, "qrinfo%c%ld%c", 0, qr.QRinfo, 0);
		fprintf(exfp, "qrfloor%c%d%c", 0, qr.QRfloor, 0);
		fprintf(exfp, "qrnumber%c%ld%c", 0, roomnums[a], 0);

		sprintf(frname, "rooms/fullrm%d", a);
		frfp=fopen(frname, "rb");
		fread(&fr, sizeof(struct fullroom), 1, frfp);
		fclose(frfp);
		if (qr.QRflags & QR_INUSE)
		    for (b=0; b<MSGSPERRM; ++b) if (fr.FRnum[b]>=mm.MMlowest) {
			++msgcount;
			fflush(stdout);
			fprintf(exfp, "message%c%ld%c", 0, fr.FRnum[b], 0);
			dump_message(exfp, fr.FRnum[b], fr.FRpos[b]);
			}

		fprintf(exfp, "endroom%c", 0);
		printf("(%d messages) ", msgcount);
		}
	    printf("\n");
	    }

	fclose(qrfp);


	/* Now do the mailboxes, which are implemented as rooms in Citadel 6 */
	usfp = fopen("usersupp", "rb+");
	if (usfp == NULL) {
		perror("Cannot open usersupp");
		fclose(exfp);
		exit(1);
		}

	printf("\n\nMailboxes\n\n");

	while (fread(&us, sizeof(struct usersupp), 1, usfp) > 0)
	  if (us.axlevel > 0) {
		msgcount = 0;
		mailroomnum = highest_roomnum++;
		printf("Mailbox for <%s> <%010ld.Mail> ",
			us.fullname, us.eternal);
		fprintf(exfp, "room%c", 0);
		fprintf(exfp, "qrname%c%010ld.Mail%c", 0, us.eternal, 0);
		fprintf(exfp, "qrpasswd%c %c", 0, 0);
		fprintf(exfp, "qrnumber%c%ld%c", 0, mailroomnum, 0);
		fprintf(exfp, "qrroomaide%c0%c", 0, 0);
		fprintf(exfp, "qrhighest%c%ld%c", 0,
				us.mailpos[MAILSLOTS-1], 0);

		/* Generation of Cit5 mail rooms will always be 1 */
		fprintf(exfp, "qrgen%c1%c", 0, 0);

		/* Fudge this using flags from Cit6: QR_INUSE and QR_MAILBOX */
		fprintf(exfp, "qrflags%c16386%c", 0, 0);

		fprintf(exfp, "qrdirname%c %c", 0, 0);
		fprintf(exfp, "qrinfo%c0%c", 0, 0);
		fprintf(exfp, "qrfloor%c1%c", 0, 0);

		for (b=0; b<MAILSLOTS; ++b) if (us.mailnum[b]>=mm.MMlowest) {
			++msgcount;
			fprintf(exfp, "message%c%ld%c", 0, us.mailnum[b], 0);
			dump_message(exfp, us.mailnum[b], us.mailpos[b]);
			}

		fprintf(exfp, "endroom%c", 0);
		printf("(%d messages) ", msgcount);
		printf("\n");
		}

	fclose(usfp);

	fprintf(exfp, "endsection%c", 0);
	}


main(int argc, char *argv[]) {
	FILE *exfp, *fp, *vifp;
	long highest_user;
	char viname[256];
	int a;

	if (argc != 3) {
		fprintf(stderr,
			"%s: usage: %s <bbs directory> <exported_filename>\n",
			argv[0], argv[0], argv[1]);
		exit(1);
		}
	
	get_config();

	if (chdir(argv[1])!=0) {
		fprintf(stderr, "%s: cannot change directory to %s: %s\n",
			argv[0], argv[1], strerror(errno));
		}
	exfp = fopen(argv[2], "wb");
	if (exfp == NULL) {
		fprintf(stderr, "%s: cannot open %s: %s\n",
			argv[0], argv[2], strerror(errno));
		exit(errno);
		}
	sprintf(viname, "%s.visits", argv[2]);
	vifp = fopen(viname, "wb");

	/* structure size variables */
	printf("\n\nStructure size variables\n\n");
	fprintf(exfp, "ssv%c", 0);
	fprintf(exfp, "maxfloors%c%d%c", 0, MAXFLOORS, 0);
	fprintf(exfp, "endsection%c", 0);

	/* Write out the server config */
	printf("\n\nServer config\n\n");
	fprintf(exfp, "config%c", 0);
	fprintf(exfp, "c_nodename%c%s%c", 0, config.c_nodename, 0);
	fprintf(exfp, "c_fqdn%c%s%c", 0, config.c_fqdn, 0);
	fprintf(exfp, "c_humannode%c%s%c", 0, config.c_humannode, 0);
	fprintf(exfp, "c_phonenum%c%s%c", 0, config.c_phonenum, 0);
	fprintf(exfp, "c_bbsuid%c%d%c", 0, config.c_bbsuid, 0);
	fprintf(exfp, "c_pwcrypt%c%d%c", 0, config.c_pwcrypt, 0);
	fprintf(exfp, "c_creataide%c%d%c", 0, config.c_creataide, 0);
	fprintf(exfp, "c_sleeping%c%d%c", 0, config.c_sleeping, 0);
	fprintf(exfp, "c_initax%c%d%c", 0, config.c_initax, 0);
	fprintf(exfp, "c_regiscall%c%d%c", 0, config.c_regiscall, 0);
	fprintf(exfp, "c_twitdetect%c%d%c", 0, config.c_twitdetect, 0);
	fprintf(exfp, "c_twitroom%c%s%c", 0, config.c_twitroom, 0);
	fprintf(exfp, "c_defent%c%d%c", 0, config.c_defent, 0);
	fprintf(exfp, "c_moreprompt%c%s%c", 0, config.c_moreprompt, 0);
	fprintf(exfp, "c_restrict%c%d%c", 0, config.c_restrict, 0);
	fprintf(exfp, "c_bbs_city%c%s%c", 0, config.c_bbs_city, 0);
	fprintf(exfp, "c_sysadm%c%s%c", 0, config.c_sysadm, 0);
	fprintf(exfp, "c_bucket_dir%c%s%c", 0, config.c_bucket_dir, 0);
	fprintf(exfp, "c_setup_level%c%d%c", 0, config.c_setup_level, 0);
	fprintf(exfp, "c_maxsessions%c%d%c", 0, config.c_maxsessions, 0);
	fprintf(exfp, "c_net_password%c%s%c", 0, config.c_net_password, 0);
	fprintf(exfp, "c_port_number%c%d%c", 0, config.c_port_number, 0);
	fprintf(exfp, "endsection%c", 0);

	/* Now some global stuff */
	printf("\n\nGlobals\n\n");
	fp = fopen("MMstructure", "rb");
	if (fp == NULL) {
		perror("Cannot open 'MMstructure' file");
		fclose(exfp);
		exit(1);
		}
	fread(&mm, sizeof(struct msgmain), 1, fp);
	fclose(fp);

	export_rooms(exfp);
	export_floors(exfp);
	export_usersupp(exfp, vifp);

	fclose(vifp);

	fprintf(exfp, "visits%c", 0);
	vifp = fopen(viname, "rb");
	while (a=getc(vifp), a>=0) putc(a, exfp);
	fclose(vifp);
	unlink(viname);
	fprintf(exfp, "endsection%c", 0);

	fprintf(exfp, "globals%c", 0);
	fprintf(exfp, "mmhighest%c%ld%c", 0, mm.MMhighest, 0);

	fp = fopen("eternal", "rb");
	if (fp == NULL) {
		perror("Cannot find highest user number");
		fclose(exfp);
		exit(1);
		}

	fread(&highest_user, sizeof(long), 1, fp);
	fclose(fp);

	fprintf(exfp, "mmnextuser%c%ld%c", 0, highest_user, 0);
	fprintf(exfp, "mmnextroom%c%ld%c", 0, highest_roomnum+1L, 0);
	fprintf(exfp, "endsection%c", 0);

	fprintf(exfp, "endfile%c", 0);
	exit(fclose(exfp));
	}
